/* 'CParticleSystem.cpp'
 * 
 * Creation:		05-07-2005
 * Last Change:		06-07-2005
 * Description:		Creates a particle system to use for special effects
 *
 */

/* include complement header file */
#include "particles/CParticleSystem.h"
#define err32 int

/* Default constructor */
CParticleSystem::CParticleSystem (void)
{
	m_pParticleBuffer  = NULL;
	m_kParticleList    = NULL;
	m_iMaxParticles    = 0;
	m_iNumParticles    = 0;
	m_fAccumulatedTime = 0.0f;
}

/* Default destructor */
CParticleSystem::~CParticleSystem (void)
{
	/* Release memory */
	if (m_kParticleList != NULL)
	{
		delete [] m_kParticleList;
		m_kParticleList = NULL;
	}

	if(m_pParticleBuffer != NULL)
	{
		m_pParticleBuffer->Release();
		m_pParticleBuffer  = NULL;
	}

	m_kParticleList = NULL;
}

/* Create new particle system */
err32 CParticleSystem::Create (u_int32 iMaxParticles, D3DXVECTOR3 vOrigin, D3DXVECTOR3 vForce, LPCSTR lpszTextureName)
{
	/* Set attributes */
	m_iMaxParticles = iMaxParticles;
	m_vOrigin		= vOrigin;
	m_vForce		= vForce;

	/* Create vertex buffer */
	if(FAILED(CScreen::GetSingleton()->GetDevice()->CreateVertexBuffer(iMaxParticles * sizeof(PointSprite),
                                            D3DUSAGE_DYNAMIC | D3DUSAGE_WRITEONLY | D3DUSAGE_POINTS, 
											D3DFVF_ParticleVertex,
                                            D3DPOOL_DEFAULT, &m_pParticleBuffer, NULL)))
    {
		return -1; //errorCreateBuffer;
    }

	/* Create texture */
	//err32 error = m_kParticleTexture.CreateFromFile(lpszTextureName);
	
	if(error != noError)
	{
		return error;
	}

	/* Allocate space for particle list */
	m_kParticleList = new Particle[m_iMaxParticles];

	/* Successful */
	return noError;
}

/* Initialize particles */
void CParticleSystem::Init (ParticleInfo * pParticleInfo)
{
	/* Initialize particle list */
	m_pParticleInfo = pParticleInfo;
}

/* Update existing particles, possibly killing or generating them */
void CParticleSystem::Update (real32 fTimePassed)
{
	/* Calculate current force */
	D3DXVECTOR3 vCurrentForce = m_vForce + ((float)rand()/RAND_MAX) * m_pParticleInfo->m_vForceVariation;

	/* Loop through all particles */
	for (u_int32 i=0; i < m_iNumParticles; )
	{
		/* Calculate current weight */
		m_kParticleList[i].m_fWeight = m_kParticleList[i].m_fStartWeight + (m_kParticleList[i].m_fEnergy/m_kParticleList[i].m_fTotalEnergy) * m_kParticleList[i].m_fWeightDelta;

		/* Calculate new velocity */
		m_kParticleList[i].m_vVelocity += vCurrentForce * m_kParticleList[i].m_fWeight;

		/* Update the particle position */
		m_kParticleList[i].m_vPosition = m_kParticleList[i].m_vPosition + (m_kParticleList[i].m_vVelocity * fTimePassed);

		/* Check for reasons to kill it */
		m_kParticleList[i].m_fEnergy -= fTimePassed;
		if(m_kParticleList[i].m_fEnergy <= 0.0)
		{
			/* Move last particle to current position */
			m_kParticleList[i] = m_kParticleList[m_iNumParticles-1];

			/* Decrease nr of particles */
			m_iNumParticles--;
		}
		else
		{
			i++;
		}
	}

	/* Store accumulated time */
	m_fAccumulatedTime += fTimePassed;

	/* Calculate nr of particles needed */
	u_int32 iNewParticles = (u_int32) ((float)m_pParticleInfo->m_iParticlesPerSecond * m_fAccumulatedTime);

	/* Save remaining time */
	m_fAccumulatedTime -= 1.0f / (float) m_pParticleInfo->m_iParticlesPerSecond * iNewParticles;

	/* Emit new particles */
	Emit(iNewParticles);
}

/* Create a couple of new particles */
int CParticleSystem::Emit (u_int32 iNrOfParticles)
{
	while (iNrOfParticles != 0 && m_iNumParticles < m_iMaxParticles)
	{
		/* Initialize new particle */
		InitializeParticle(m_iNumParticles);

		/* Increase number of particles */
		m_iNumParticles++;

		/* Decrease nr of particles to create */
		iNrOfParticles--;
	}

	/* Return nr of uncreated particles */
	return iNrOfParticles;
}

/* Particle initialization */
void CParticleSystem::InitializeParticle (u_int32 index)
{
	/* Determine starting position */
	m_kParticleList[index].m_vPosition.x = m_vOrigin.x + ((float)rand()/RAND_MAX) * m_pParticleInfo->m_vPositionVariation.x;
	m_kParticleList[index].m_vPosition.y = m_vOrigin.y + ((float)rand()/RAND_MAX) * m_pParticleInfo->m_vPositionVariation.y;
	m_kParticleList[index].m_vPosition.z = m_vOrigin.z + ((float)rand()/RAND_MAX) * m_pParticleInfo->m_vPositionVariation.z;

	/* Set the velocity */
	m_kParticleList[index].m_vVelocity.x = m_pParticleInfo->m_vStandardVelocity.x + ((float)rand()/RAND_MAX) * m_pParticleInfo->m_vVelocityVariation.x;
	m_kParticleList[index].m_vVelocity.y = m_pParticleInfo->m_vStandardVelocity.y + ((float)rand()/RAND_MAX) * m_pParticleInfo->m_vVelocityVariation.y;
	m_kParticleList[index].m_vVelocity.z = m_pParticleInfo->m_vStandardVelocity.z + ((float)rand()/RAND_MAX) * m_pParticleInfo->m_vVelocityVariation.z;

	/* Set size */
	m_kParticleList[index].m_fSize = m_pParticleInfo->m_fSize;

	/* Set weight */
	m_kParticleList[index].m_fStartWeight = m_kParticleList[index].m_fWeight = m_pParticleInfo->m_fStandardWeight + ((float)rand()/RAND_MAX) * m_pParticleInfo->m_fWeightVariation;
	m_kParticleList[index].m_fWeightDelta = m_pParticleInfo->m_fStandardWeightDelta + ((float)rand()/RAND_MAX) * m_pParticleInfo->m_fWeightDeltaVariation;

	/* Set color */
	m_kParticleList[index].m_cStartColor.a = m_pParticleInfo->m_cStandardStartColor.a + ((float)rand()/RAND_MAX) * m_pParticleInfo->m_cStartColorVariation.a;
	m_kParticleList[index].m_cStartColor.r = m_pParticleInfo->m_cStandardStartColor.r + ((float)rand()/RAND_MAX) * m_pParticleInfo->m_cStartColorVariation.r;
	m_kParticleList[index].m_cStartColor.g = m_pParticleInfo->m_cStandardStartColor.g + ((float)rand()/RAND_MAX) * m_pParticleInfo->m_cStartColorVariation.g;
	m_kParticleList[index].m_cStartColor.b = m_pParticleInfo->m_cStandardStartColor.b + ((float)rand()/RAND_MAX) * m_pParticleInfo->m_cStartColorVariation.b;

	m_kParticleList[index].m_cEndColor.a = m_pParticleInfo->m_cStandardEndColor.a + ((float)rand()/RAND_MAX) * m_pParticleInfo->m_cEndColorVariation.a;
	m_kParticleList[index].m_cEndColor.r = m_pParticleInfo->m_cStandardEndColor.r + ((float)rand()/RAND_MAX) * m_pParticleInfo->m_cEndColorVariation.r;
	m_kParticleList[index].m_cEndColor.g = m_pParticleInfo->m_cStandardEndColor.g + ((float)rand()/RAND_MAX) * m_pParticleInfo->m_cEndColorVariation.g;
	m_kParticleList[index].m_cEndColor.b = m_pParticleInfo->m_cStandardEndColor.b + ((float)rand()/RAND_MAX) * m_pParticleInfo->m_cEndColorVariation.b;
	
	/* Set energy */
	m_kParticleList[index].m_fTotalEnergy = m_kParticleList[index].m_fEnergy = m_pParticleInfo->m_fStandardEnergy + ((float)rand()/RAND_MAX) * m_pParticleInfo->m_fEnergyVariation;
}

/* Render the particles */
err32 CParticleSystem::Render (void)
{
	/* Check if there is anything to render first */
	if (m_iNumParticles == 0)
	{
		return noError;
	}

	/* Create point sprites */
	PointSprite *sprites = new PointSprite[m_iNumParticles];

	/* Set sprite properties */
	for (u_int32 i=0; i < m_iNumParticles; i++)
	{
		/* Set position */
		sprites[i].x = m_kParticleList[i].m_vPosition.x;
		sprites[i].y = m_kParticleList[i].m_vPosition.y;
		sprites[i].z = m_kParticleList[i].m_vPosition.z;

		/* Set color */
		sprites[i].color = m_kParticleList[i].m_cEndColor - ((m_kParticleList[i].m_fEnergy/m_kParticleList[i].m_fTotalEnergy) * (m_kParticleList[i].m_cEndColor - m_kParticleList[i].m_cStartColor));
	}

	/* Fill vertex buffer */
	PointSprite *ptr;

	/* Lock the vertex buffer */
	if(FAILED(m_pParticleBuffer->Lock(0, m_iNumParticles * sizeof(PointSprite), (void**)&ptr, 0)))
	{
		return -1;//errorLockBuffer;
	}

	/* Here we copy the vertex data into the vertex buffer */
	memcpy(ptr, sprites, m_iNumParticles * sizeof(PointSprite));

	/* Unlock when done copying data into the buffer. */
	m_pParticleBuffer->Unlock();

	/* free memory */
	delete [] sprites;

	/* Set texture */
	m_kParticleTexture.SetActiveTexture();

	/* Set render states */
	CScreen::GetSingleton()->GetDevice()->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
	CScreen::GetSingleton()->GetDevice()->SetRenderState(D3DRS_ZENABLE, D3DZB_FALSE);
	CScreen::GetSingleton()->GetDevice()->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
    CScreen::GetSingleton()->GetDevice()->SetRenderState(D3DRS_POINTSPRITEENABLE, TRUE);
    CScreen::GetSingleton()->GetDevice()->SetRenderState(D3DRS_POINTSCALEENABLE, TRUE);
    CScreen::GetSingleton()->GetDevice()->SetRenderState(D3DRS_POINTSIZE, FtoDW(m_kParticleList[m_iNumParticles-1].m_fSize));
    CScreen::GetSingleton()->GetDevice()->SetRenderState(D3DRS_POINTSIZE_MIN, FtoDW(1.0f));
    CScreen::GetSingleton()->GetDevice()->SetRenderState(D3DRS_POINTSCALE_A, FtoDW(0.0f));
    CScreen::GetSingleton()->GetDevice()->SetRenderState(D3DRS_POINTSCALE_B, FtoDW(0.0f));
    CScreen::GetSingleton()->GetDevice()->SetRenderState(D3DRS_POINTSCALE_C, FtoDW(1.0f));

	/* Draw point sprites */
	CScreen::GetSingleton()->GetDevice()->SetStreamSource(0, m_pParticleBuffer, 0, sizeof(PointSprite));
    CScreen::GetSingleton()->GetDevice()->SetFVF(D3DFVF_ParticleVertex);
    CScreen::GetSingleton()->GetDevice()->DrawPrimitive(D3DPT_POINTLIST, 0, m_iNumParticles);

	/* Restore render states */
    CScreen::GetSingleton()->GetDevice()->SetRenderState(D3DRS_POINTSPRITEENABLE, FALSE);
    CScreen::GetSingleton()->GetDevice()->SetRenderState(D3DRS_POINTSCALEENABLE, FALSE);
    CScreen::GetSingleton()->GetDevice()->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
	CScreen::GetSingleton()->GetDevice()->SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);

	return noError;
}

/* Reset the particle system */
void CParticleSystem::Reset (void)
{
	m_iNumParticles = 0;
}

/* Move the particle system */
void CParticleSystem::Move (D3DXVECTOR3 vDirection)
{
	m_vOrigin += vDirection;
}

/* Set new origin */
void CParticleSystem::SetOrigin (D3DXVECTOR3 vOrigin)
{
	m_vOrigin = vOrigin;
}

/* Change force */
void CParticleSystem::SetForce (D3DXVECTOR3 vForce)
{
	m_vForce = vForce;
}

/* Get current origin */
D3DXVECTOR3 CParticleSystem::GetOrigin (void)
{
	return m_vOrigin;
}

/* Get current force */
D3DXVECTOR3 CParticleSystem::GetForce (void)
{
	return m_vForce;
}

/* Get current origin */
u_int32 CParticleSystem::GetNrOfParticles (void)
{
	return m_iNumParticles;
}