#include <hush/source.h>

#include <hush/flags.h> // to check for disabled lists




#ifndef NOLIST

#include <hush/list.h>
#include <hush/handler.h>
#include <hush/garbage.h>

#include <hush/util/baselist.h>
#include <hush/sys/cell.h>
#include <hush/assert.h>

#include "hush/sys/local.i"
#include "hush/sys/macros.i"

#define TYPE dvlist_type
#include "hush/sys/table.i"

//// Abbreviations for names and access

#define ME ((list<void>*)_me)

#define INSERTION ((baselist::element_void) ME->insertion())
#define EXTRACTION ((baselist::element_void) ME->extraction())
#define EQUALITY ((baselist::compare_void) ME->equality())
#define ORDER ((baselist::compare_void) ME->order())
#define FORMAT ((baselist::format_void) ME->format())
#define RESULT (_me?EXTRACTION(result,*_me):result)
#define LIST ((list<void>*) _list )

//// For garbage collection and statistics

//int list_type::created = 0;
//int list_type::deleted = 0;

int baselist::created = 0;
int baselist::deleted = 0;

int dvbasenode::created = 0;
int dvbasenode::deleted = 0;

int dvsublist_type::created = 0;
int dvsublist_type::deleted = 0;

int dvlistelement_type::created = 0;
int dvlistelement_type::deleted = 0;


dvlist_type::dvlist_type() { created ++; }
dvlist_type::~dvlist_type() { deleted++; }

dvsublist_type::dvsublist_type() { created ++; }
dvsublist_type::~dvsublist_type() { deleted++; }

dvlistelement_type::dvlistelement_type() { created ++; }
dvlistelement_type::~dvlistelement_type() { deleted++; }

dvbasenode::dvbasenode(const void* e, dvbasenode* n, dvbasenode* p){ 
	el = e; next=n; prev = p; created ++; 
}
dvbasenode::~dvbasenode() { deleted++; }

class HUSH_API listiter : public iter<void> {
public:
  listiter(const dvbasenode* c, const list_type* me) : iter<void>(this) {
    _current  = c; _me = me;
  } 
  void* operator()() {
    void* result = (void*) (_current?_current->el:0);
    if (_current) _current = _current->next;
    return result?RESULT:0;
  }
private:
  const dvbasenode* _current;
  const list_type* _me;
};



//// Construction 

baselist::baselist(const list_type* me) {	
  baselist::created++;
  init(me);
}

baselist::baselist(dvbasenode* x, const list_type* me) { 
  baselist::created++;
  init(me);
  _insert(x,0);
}

void baselist::init(const list_type* me) {
  _me = me;
  refs = 0;
  _len = 0; _index=0;
  _current = 0; _first = 0; _last = 0; 
  _garb = 0; _lgarb = 0; _igarb = 0; _egarb = 0; _sgarb=0;
}

baselist* baselist::clone() const{
  list<void>* tmp = new list<void>;
  tmp->_copy( *ME );
  listiter it(_first, _me);
  void* p = 0;
  while ( (p = it() ) != 0 ) tmp->insert(p);
  return (baselist*) tmp; // no registration 
}


//// The basic list operations: insert, remove, find

void baselist::insert(const void* e, int before_current) { /// insert e at right place
	//cerr << " l:i:0 " << this << " " << e << endl;
  require (_len == 0 || _current);				/// pre condition: invariant
  //cerr << " l:i:1 " << this << " " << e << " " << _me << endl;
  void* el = (void*) (_me?INSERTION(e,*_me):e);	/// test/modify insertion
  //cerr << " l:i:2 " << this << " " << e << endl;
  if (!el) return;	/// no insertion required 
  //cerr << " l:i:3 " << this << " " << e << endl;
  int before=before_current;					/// insert before head?
  //cerr << " l:i:4 " << this << " " << e << endl;
  if (_me && ME->ordered())  find(e, &before);	/// find insertion place
  else if (_me && 
	(ME->lifo() || ME->fifo())) end(); /// fifo/lifo: insert at end
  //cerr << " l:i:6 " << this << " " << e << endl;
  _insert(el, before);
  //cerr << " l:i:7 " << this << " " << e << endl;
}

// insert e after (before) current
void baselist::_insert(const void* e, int before) { 
  require (_len == 0 || _current);		/// pre condition: invariant
  dvbasenode *target = new dvbasenode(e);	/// create new node
  if (!_current)  				/// list still empty
    _first = _last = target;			/// trivial
  else {					/// the real work:
    dvbasenode *next = _current->next;		/// will be used later
    dvbasenode *prev = _current->prev;		/// will be used later
    if (_current == _first && before) {		/// insert before head
      target->prev = 0;
      target->next = _current;
      _current->prev = target;
      _first = target;
    } else if (_current == _last && !before) {	/// insert after end
      target->prev = _last;
      target->next = next;
      _current->next = target;
      _last = target;
      _index = _len;				
    } else if (!before) {			/// normal case, after current;
      target->prev = _current;
      target->next = next;
      next->prev = target;
      _current->next = target;
      _index++;
    } else { 					/// normal case, before current;
      target->next = _current;
      target->prev = prev;
      prev->next = target;
      _current->prev = target;
    }
  }
  _len++;					/// increase length;
  _current = target;				/// set new current
}

void* baselist::remove() { // remove current;
  if (!_current) return 0;
  
  dvbasenode* target = _current;
  dvbasenode* next   = target->next;
  dvbasenode* prev   = target->prev;
  
  if (prev == 0 && next == 0) { /// delete only item
    _current = _first = _last = 0; _index = 0;
  } else if (prev == 0) {	/// delete first item
    _first = _current = next;
    next->prev = 0;
  } else if (next == 0) { 	/// delete tail item
    _last = _current = prev;
    prev->next = 0;
    _index--;
  } else { 			/// normal case
    prev->next = next;
    next->prev = prev;
    _current = next;
  }
  _len--;				/// decrease length

  const void* result = target->el;	/// save element
  delete target;			/// destruct target;
  return EXTRACTION(result,*_me);	/// return target's contents
  
}

void* baselist::find(const void* target, int* before) {
  if (before) *before = 0; 
  if (_len == 0) return 0;
  
  /// maybe we are lucky:
  if ( EQUALITY(_current->el,target) ) return EXTRACTION(_current->el,*_me);
  
  /// we're not ... lets find target:
  if (!(ME->ordered())) { 	/// List not ordered... 
    first(); 			/// Do linear search from start
    do {
      if ( EQUALITY(_current->el,target) ) return EXTRACTION(_current->el,*_me);
    } while( next() );
    return 0;			/// not found ...
    
  } else {			/// list ordered. 
    if ( ORDER(_last->el, target) ) {	
      end(); return 0;		/// not found ...
    }
    if ( ORDER(_current->el,target) ) { /// current < target 
      do { 				/// Do linear search up from current
        if (EQUALITY(_current->el,target)) return EXTRACTION(_current->el,*_me);
      } while(ORDER(_current->el,target) && next());
      _current = _current->prev; 	/// adjust current to meet precondition
      _index--;
      return 0; 			/// not found
      
    } else {				/// current > target
      do {				///  Do linear search down from current
        if (EQUALITY(_current->el,target)) return EXTRACTION(_current->el,*_me);
      } while (!ORDER(_current->el,target) && previous());
      
      // not found. Set "before" for insert operation:
      if (before && !ORDER(_current->el,target)) *before = 1;
      return 0;
    }
  }
}

//// Basic functions to retrieve information
//   All const

int baselist::empty()  const { return length() == 0; }
int baselist::index()  const { return _index; }
int baselist::length() const { return _len; }

void* baselist::head() const { return _first?EXTRACTION(_first->el, *_me):0;}
void* baselist::last() const { return _last ?EXTRACTION(_last->el,  *_me):0;}

int baselist::more() const { return _current?(_current->next != 0):0; }
int baselist::less() const { return _current?(_current->prev != 0):0; }

void* baselist::current() const {
  require(_current || (_len == 0) );
  if (!_current) return 0;
  return EXTRACTION(_current->el,*_me);
}

int baselist::contains(const void* x) const {
  //return 0;
  //cerr << "l:c " << x << " " << endl;
  //cerr << "l:me " << _me << endl;
  if (!_me) return 0;
  //cerr << "l:c:0 " << x  << endl;
  listiter it(_first, _me);
  //listiter it(_first,this);
  //cerr << "l:c:1 " << x << endl;
  void* p = 0; 
  while ((p = it())) if ( EQUALITY(p,x) ) return 1;
  return 0;
}

int baselist::count(const void* x) const {
  if (x == 0) return length();
  listiter it(_first, _me);
  void* p = 0; int counter = 0;
  while ((p = it())) if (EQUALITY(p,x)) counter++;
  return counter;
}

//// Basic cursor operations

void* baselist::next() { 
  if (more()) { _current = _current->next; _index++; return current();
  } else return 0;
}
void* baselist::previous() { 
  if (less()) { _current = _current->prev; _index--; return current();
  } else return 0;
}

void* baselist::first() { _current = _first; _index=0; return current(); }

void* baselist::end() { _current =  _last; _index = _len-1; return current(); }

/// a list of 5 elements a,b,c,d,e is numbered as:
/// positive index: 	 0	 1	 2	 3	 4
/// elements:		 a	 b	 c	 d	 e
/// negative index: 	-5	-4	-3	-2	-1

void* baselist::current(int index) {
  if (!_current) return 0;

  index = index % _len;		/// normalize index
  if (index < 0) index += _len;

  if (index == _len-1) return end();
  else if (index == 0) return first();
  else {
    while (index != _index) 
      if (index < _index) previous(); else next();
    return current();
  }
}

//// Misc operations

void baselist::rotate(int n) {
  const int len = length();		/// get length
  if (len < 2) return;			/// it's no use ...
  
  const int index = n % len;		/// rotate len mod n steps
  if (index == 0) return;		/// it's no use ...
  
  current(index);			/// set current to new head
  
  _last->next  = _first;		/// make list circular
  _first->prev = _last;			/// ...
  
  _first = _current;			/// set new first
  _last  = _current->prev;		/// set new last
  
  _last->next = 0;			/// break circle
  _first->prev = 0;			/// ...

  _index = 0;				/// new _index
}


void* baselist::next(const void* x) {
  require(_current || _len==0);
  if (!_current) return 0;
  while(next()) {
    const void* p = _current->el;
    if ( EQUALITY(p,x) ) return EXTRACTION(p,*_me);
  } 
  return 0;
}

void baselist::linsert(const list_type& x) {
  list<void>& l = (list<void>&) x;
  iter<void>& it = l;
  void* p = 0;
  while( (p = it()) ) insert(p);
}

void baselist::lremove(const list_type& x) {
  list<void>& l = (list<void>&) x;
  iter<void>& it = l;
  void* p = 0;
  while( (p = it()) ) if (find(p)) remove();
}

void* baselist::replace(const void* el) { // replace current el
  if (!_current) return 0;
  const void* result = _current->el;
  _current->el = el;
  return EXTRACTION(result,*_me);
}

void* baselist::replace(const void* el, int index, int len) {
  if (!_current || len <= 0) return 0;
  if (-index > length() || index >= length()) return 0;
  
  if (index < 0) index += length();	/// normalize index
  current(index);			/// adjust current
  void* result = replace(el);   	/// replace current
  
  if (len > 1) remove(index+1, len);	/// remove rest: potential memory leak!
  current(index);			/// repair current
  return EXTRACTION(result,*_me);
}

void baselist::replace(const list_type& x, int index, int len) {
  if (!_current || len <= 0) return;
  if (-index > length() || index >= length()) return;
  
  if (index < 0) index += length();	/// normalize index
  remove(index, len);			/// remove : potential memory leak!
  
  list<void>& l = (list<void>&) x;
  iter<void>& it = l;
  void* p = it();			/// first item
  _insert(p, 1);				/// insert before current
  while( (p = it()) ) _insert(p,0);	/// insert after current
  current(index);			/// repair current
}

#ifndef NOIOSTREAM
void baselist::print(ostream& os)  {
  listiter it(_first, _me); void* p = 0; 
  while (( p = it())) FORMAT(os,p);
}
#endif

void baselist::remove(int pos, int len) { 
  if (len == 0) return;
  else if (len > _len) len = _len;
  current(pos);
  if (len >= 0) 
    for(int i=0; i<len && remove(); i++);
  else {
    for(int i=0; i>len && remove() && previous(); i--);
    next(); // adjust current
  }
}

//// Processing

int baselist::process( void* f ) {  /// side-effects!
  listiter it(_first, _me); void* p = 0; 
  int tmp = 0;
  while ((p = it())) tmp += ((predicate_void)f)(p,*_me); /// do the thing
  return tmp;
}

//// Applying functions

list_type* baselist::apply( void* f ) {  /// with list
  list<void>* tmp = new list<void>;
  listiter it(_first, _me); void* p = 0; 
  while ( (p = it()) ) {
    void* e = ((apply_void)f)(p,*_me);
    if (e) tmp->insert(e);
  }
  return tmp;
}

iter<void>* baselist::iterator() const {
  iter<void>* result = new listiter(_first, _me);
  i_register(result);
  return result;
} 


// should be optimized for ordered lists:
list_type* baselist::select(const void* el) {  
  list<void>* tmp = new list<void>; 
  listiter it(_first, _me); void* p = 0; 
  while ((p = it())) if (EQUALITY(p, el)) tmp->insert(p);
  return tmp;
}


/* ------------------------------------------------------------ */

// take me higher

#ifdef recursive_templates

list_type* baselist::map( void* f ) {
  list< list<void> >* tmp = new list< list<void> >;
  listiter it(_first, _me);
  void* p = 0;
  while ( (p = it()) ) {
    list<void>* x = (list<void>*) ((mapf_void)f)(p,ME); /// do it 
    x->_copy( ME );
    //ME->_register(x); // register for removal
    tmp->insert( x ); 
  }
  //ME->_register( (list<void>*) tmp);
  return tmp;
}  // deletion is done automatically

#endif // no_recursive_templates

void baselist::_register(void* p)  const { REGISTER(((baselist*)this)->_garb  ,p); }
void baselist::l_register(void* p) const { REGISTER(((baselist*)this)->_lgarb ,p); }
void baselist::i_register(void* p) const { REGISTER(((baselist*)this)->_igarb ,p); }
void baselist::s_register(void* p) const { REGISTER(((baselist*)this)->_sgarb ,p); }
void baselist::e_register(void* p) const { REGISTER(((baselist*)this)->_egarb ,p); }


baselist::~baselist() {
  baselist::deleted++;
  HUSHDELETE(_garb,garbage<void>);
  HUSHDELETE(_lgarb,list<void>);
  HUSHDELETE(_igarb,iter<void>);
  HUSHDELETE(_egarb,dvlistelement<void>);
  HUSHDELETE(_sgarb,dvsublist<void>);
} 

#endif // NOLIST
