//-------------------------------------------------------------------

float4x4 WorldViewProj : WorldViewProj;
float4   Dimensions;
texture  DiffuseMap;

//-------------------------------------------------------------------

sampler DiffuseMapSampler = sampler_state 
{
    Texture   = <DiffuseMap>;
    MinFilter = POINT;
    MagFilter = POINT;
    MipFilter = NONE;
    AddressU  = WRAP;
    AddressV  = WRAP;
};

//-------------------------------------------------------------------

struct VS_INPUT 
{
    float4 Position : POSITION;  //in object space
    float3 Normal   : NORMAL;    //in object space
    float2 TexCoord : TEXCOORD0;
    float3 T        : TEXCOORD1; //in object space
    float3 B        : TEXCOORD2; //in object space
    float3 N        : TEXCOORD3; //in object space
};

struct VS_OUTPUT 
{
    float4 Position  : POSITION;  //in projection space
    float4 TexCoord0 : TEXCOORD0;
    float4 TexCoord1 : TEXCOORD1;
    float4 TexCoord2 : TEXCOORD2;
    float4 TexCoord3 : TEXCOORD3;
};

//-------------------------------------------------------------------
// VERTEX SHADERS

VS_OUTPUT SimpleVS(VS_INPUT IN)
{
    VS_OUTPUT OUT;

    // transform position to projection space
    OUT.Position = mul(IN.Position, WorldViewProj);

    // pass texture coordinates for fetching the diffuse map
    // t0 holds addresses for top-left     quadrant 
    // t1 holds addresses for top-right    quadrant 
    // t2 holds addresses for bottom-left  quadrant 
    // t3 holds addresses for bottom-right quadrant 
    float4 kMultipliers = float4(-1.0f, 0.0f, 1.0f, 2.0f);
    OUT.TexCoord0.xyzw = kMultipliers.xxyy * Dimensions.xyxy + IN.TexCoord.xyxy;
    OUT.TexCoord1.xyzw = kMultipliers.zxwy * Dimensions.xyxy + IN.TexCoord.xyxy;
    OUT.TexCoord2.xyzw = kMultipliers.xzyw * Dimensions.xyxy + IN.TexCoord.xyxy;
    OUT.TexCoord3.xyzw = kMultipliers.zzww * Dimensions.xyxy + IN.TexCoord.xyxy;

     return OUT;
}

//-------------------------------------------------------------------
// PIXEL SHADER HELPER FUNCTIONS

float4 ComputeWeights(const float dist)
{
    float4 d        = dist * float4(1,1,-1,-1) + float4(1.0,0.0,1.0,2.0); // (1.0+weight, 0.0+weight, 1.0-weight, 2.0-weight)
    float4 dSquared = d*d;
    float4 dCubed   = dSquared * d;

    // A = -0.75
    float4 h =      float4( 3.0,     1,    1,    3);
    h += d        * float4(-6.0,     0,    0,   -6);
    h += dSquared * float4( 3.75,-2.25,-2.25, 3.75);
    h += dCubed   * float4(-0.75, 1.25, 1.25,-0.75);

    return h;
}

half4 cubicFilter(const float4 h, const half4 c0, const half4 c1, const half4 c2, const half4 c3)
{
    half4 r;

    r  = c0 * h.x;
    r += c1 * h.y;
    r += c2 * h.z;
    r += c3 * h.w;

    return r;
}

//-------------------------------------------------------------------
// PIXEL SHADERS

float4 PointSamplePS(VS_OUTPUT IN) : COLOR
{
    //fetch base color
    float2 addr  = IN.TexCoord0.zw + 0.5f * Dimensions.xy;
    half4 color = tex2D(DiffuseMapSampler, addr);

    return color;
}

float4 BilinearPS(VS_OUTPUT IN) : COLOR
{
    half2 oneZero  = half2(1.0f, 0.0f);
    float4 addrLRTB = IN.TexCoord0.zzww + oneZero.xyxy * Dimensions.xxyy;
    float2 weightLT = frac(addrLRTB.xz * Dimensions.zw);
    
    half4 topleftC     = tex2D(DiffuseMapSampler, addrLRTB.xz);
    half4 toprightC    = tex2D(DiffuseMapSampler, addrLRTB.yz);
    half4 bottomleftC  = tex2D(DiffuseMapSampler, addrLRTB.xw);
    half4 bottomrightC = tex2D(DiffuseMapSampler, addrLRTB.yw);

    half4 top    = lerp(toprightC,    topleftC,    weightLT.xxxx);
    half4 bottom = lerp(bottomrightC, bottomleftC, weightLT.xxxx);

    half4 color = lerp(bottom, top, weightLT.yyyy);

    return color;
}

float4 BicubicPS(VS_OUTPUT IN) : COLOR
{
    float2 weightLT  = frac(IN.TexCoord0.zw * Dimensions.zw);
  
    float4 h = ComputeWeights(weightLT.x);
    half4 row0 = cubicFilter(h, 
                              tex2D(DiffuseMapSampler, IN.TexCoord0.xy),
                              tex2D(DiffuseMapSampler, IN.TexCoord0.zy),
                              tex2D(DiffuseMapSampler, IN.TexCoord1.xy),
                              tex2D(DiffuseMapSampler, IN.TexCoord1.zy));
    half4 row1 = cubicFilter(h, 
                              tex2D(DiffuseMapSampler, IN.TexCoord0.xw),
                              tex2D(DiffuseMapSampler, IN.TexCoord0.zw),
                              tex2D(DiffuseMapSampler, IN.TexCoord1.xw),
                              tex2D(DiffuseMapSampler, IN.TexCoord1.zw));
    half4 row2 = cubicFilter(h, 
                              tex2D(DiffuseMapSampler, IN.TexCoord2.xy),
                              tex2D(DiffuseMapSampler, IN.TexCoord2.zy),
                              tex2D(DiffuseMapSampler, IN.TexCoord3.xy),
                              tex2D(DiffuseMapSampler, IN.TexCoord3.zy));
    half4 row3 = cubicFilter(h, 
                              tex2D(DiffuseMapSampler, IN.TexCoord2.xw),
                              tex2D(DiffuseMapSampler, IN.TexCoord2.zw),
                              tex2D(DiffuseMapSampler, IN.TexCoord3.xw),
                              tex2D(DiffuseMapSampler, IN.TexCoord3.zw));

    h = ComputeWeights(weightLT.y);
    half4 color = cubicFilter(h, row0, row1, row2, row3);

    return color;
}

//-------------------------------------------------------------------
// TECHNIQUES

Technique PointSample
{
    Pass P0
    {
        VertexShader = compile vs_1_1 SimpleVS();
        PixelShader  = compile ps_2_0 PointSamplePS();
        CullMode = None;
    }
}

Technique Bilinear
{
    Pass P0
    {
        VertexShader = compile vs_1_1 SimpleVS();
        PixelShader  = compile ps_2_0 BilinearPS();
        CullMode = None;
    }
}

Technique Bicubic
{
    Pass P0
    {
        VertexShader = compile vs_1_1 SimpleVS();
        PixelShader  = compile ps_2_0 BicubicPS();
        CullMode = None;
    }
}

//-------------------------------------------------------------------
